using System;
using System.Collections;
using System.ComponentModel;
using Flock;
using Vector;
 // === File Prolog ============================================================
 // This code was developed by RiverSoftAVG.
 // --- Notes ------------------------------------------------------------------
 // --- Development History  ---------------------------------------------------
 // 09/2004 T. Grubb
 // - Deleted AfterConstruction method and put code in constructor
 // (required by Delphi.NET)
 // 03/2001	   T. Grubb
 // Initial version.
 // Known Bugs:
 // - TFormationFlock.Execute command does not interpolate properly
 // when rotating to new desiredvelocity.  It oscillates around
 // desired velocity
 // File Contents:
 // TFormationBoid        One Boid (Member) of a flock that can be in
 // formation
 // TFormationFlock       Flock of boids in formation
 // TFormationFlockCollection Collection class to hold formation boids
 // --- Warning ----------------------------------------------------------------
 // This software is property of RiverSoftAVG. Unauthorized use or
 // duplication of this software is strictly prohibited. Authorized users
 // are subject to the following restrictions:
 // *	RiverSoftAVG is not responsible for
 // any consequence of the use of this software.
 // *	The origin of this software must not be misrepresented either by
 // explicit claim or by omission.
 // *	Altered versions of this software must be plainly marked as such.
 // *	This notice may not be removed or altered.
 // This software is freeware.  You are authorized to duplicate and modify
 // this software subject to the restrictions above.
 //  2001-2007, Thomas G. Grubb
 // === End File Prolog ========================================================
namespace FormationFlock
{
    public class TFormationBoid: TBoid
    {
        public double DistanceSqrFromOffset
        {
          get {
            return FDistanceSqrFromOffset;
          }
        }
        // Published declarations
        public TVector Offset
        {
          get {
            return FOffset;
          }
          set {
            SetOffset(value);
          }
        }
        public TVector OrientationOffset
        {
          get {
            return FOrientationOffset;
          }
          set {
            SetOrientationOffset(value);
          }
        }
        // Private declarations
        private TVector FOffset = null;
        private TVector FOrientationOffset = null;
        private double FDistanceSqrFromOffset = 0;
        // Protected declarations
        // TFormationBoid
        public override void ApplyCohesion(TVector Accumulator)
        {
            TVector Change;
            // Flocking Rule: Cohesion
            // Try to go toward our offset position from the flock's center position
            // Copy the position of center of flock
            //@ Unsupported property or method(C): 'Position'
            Change = new TVector(Flock.Position);
            try {
                // Offset
                Change.Add(Offset);
                // Calculate the distance to our desired position and save it for other routines
                FDistanceSqrFromOffset = Change.DistanceSqrTo(this.Position);
                Change.Subtract(this.Position);
                // normalize change to look more natural
                //@ Unsupported property or method(C): 'MaxUrgency'
                //@ Unsupported property or method(C): 'StrengthCohesion'
                Change.SetMagnitudeOfVector(Flock.MaxUrgency * Flock.StrengthCohesion);
                // Add Change to Accumulator
                Accumulator.Add(Change);
            } finally {
                //@ Unsupported property or method(C): 'Free'
                Change.Free;
            }
        }

        public override void Assign(Object Source)
        {
            if (Source is TFormationBoid)
            {
                FOffset.Assign(((Source) as TFormationBoid).Offset);
                FOrientationOffset.Assign(((Source) as TFormationBoid).OrientationOffset);
                base.Assign(Source);
            }
            else
            {
                base.Assign(Source);
            }
        }

        // Public declarations
        //Constructor  Create( Collection)
        public TFormationBoid(ArrayList Collection) : base(Collection)
        {
            FOffset = new TVector();
            FOrientationOffset = new TVector();
            if (Flock is TFormationFlock)
            {
                ((Flock) as TFormationFlock).FormationChange();
            }
        }
        //@ Destructor  Destroy()
        ~TFormationBoid()
        {
            TFormationFlock FormationFlock;
            if (Flock is TFormationFlock)
            {
                FormationFlock = ((Flock) as TFormationFlock);
            }
            else
            {
                FormationFlock = null;
            }
            //@ Unsupported property or method(C): 'Free'
            FOffset.Free;
            //@ Unsupported property or method(C): 'Free'
            FOrientationOffset.Free;
            base.Destroy();
            if (FormationFlock != null)
            {
                FormationFlock.FormationChange();
            }
        }
        private void SetOffset(TVector Value)
        {
            FOffset.Assign(Value);
            //@ Undeclared identifier(3): 'Changed'
            Changed(false);
        }

        private void SetOrientationOffset(TVector Value)
        {
            FOrientationOffset.Assign(Value);
            //@ Undeclared identifier(3): 'Changed'
            Changed(false);
        }

        public override void Execute()
        {
            // The update method is where all the actual flocking is done.  It is called
            // every time period and the boid updates its position based on its flocking
            // behaviors
            base.Execute();
            // if we are close to our position then stop
            //@ Unsupported property or method(C): 'MaxChange'
            //@ Unsupported property or method(C): 'MaxChange'
            if (DistanceSqrFromOffset < (Flock.MaxChange * Flock.MaxChange))
            {
                this.Speed = 0;
                this.Velocity.SetMagnitudeOfVector(this.Speed);
            }
        }

        public override void UpdateVelocity(TVector Accumulator)
        {
            float Dist;
            base.UpdateVelocity(Accumulator);
            // if we are almost to our destination, start slowing down
            Dist = Math.Sqrt(FDistanceSqrFromOffset);
            //@ Unsupported property or method(C): 'MaxChange'
            //@ Unsupported property or method(C): 'MinTimeToMaxSpeed'
            if ((Dist / Flock.MaxChange) < Flock.MinTimeToMaxSpeed)
            {
                //@ Unsupported property or method(C): 'MaxSpeed'
                //@ Unsupported property or method(C): 'MaxChange'
                this.Velocity.Multiply(Flock.MaxSpeed - (Flock.MaxChange / Dist));
            }
        }

        public override void ApplySeparation(TVector Accumulator)
        {
            float Ratio;
            TVector Change;
            // we recode this rule to make sure we don't move away when we are in formation
            //@ Unsupported property or method(C): 'StrengthSeparation'
            if (Flock.StrengthSeparation == 0)
            {
                return;
            }
            // Flocking Rule: Separation
            // Try to avoid boid colliding with its flockmates
            Change = new TVector(this.NearestFlockMate.Position);
            try {
                // compute vector towards our nearest flock mate
                Change.Subtract(this.Position);
                // Compute ratio of nearest flock mate to separation distance
                // and ensure it is in range of MinUrgency..MaxUrgency
                //@ Unsupported property or method(C): 'SeparationDistance'
                Ratio = this.NearestFlockMateDist / Flock.SeparationDistance;
                //@ Unsupported property or method(C): 'MinUrgency'
                if (Ratio < Flock.MinUrgency)
                {
                    //@ Unsupported property or method(C): 'MinUrgency'
                    Ratio = Flock.MinUrgency;
                }
                //@ Unsupported property or method(C): 'MaxUrgency'
                else if (Ratio > Flock.MaxUrgency)
                {
                    //@ Unsupported property or method(C): 'MaxUrgency'
                    Ratio = Flock.MaxUrgency;
                }
                //@ Unsupported property or method(C): 'StrengthSeparation'
                Ratio = Ratio * Flock.StrengthSeparation;
                // Are we too close to nearest flockmate?  Then Move Away
                //@ Unsupported property or method(C): 'SeparationDistance'
                if (this.NearestFlockMateDist < Flock.SeparationDistance * 1.5)
                {
                    // Are we too far from nearest flockmate?  Then Move Closer
                    Change.SetMagnitudeOfVector( -Ratio);
                }
                //@ Unsupported property or method(C): 'SeparationDistance'
                else if (this.NearestFlockMateDist > Flock.SeparationDistance * 1.5)
                {
                    Change.SetMagnitudeOfVector(Ratio);
                }
                else
                {
                    // just right
                    Change.Clear();
                }
                // Add Change to Accumulator
                if (Flock != null)
                {
                    ((Flock) as TFormationFlock).BoidApplyRule(this, Flock.TFlockingBehavior.fbSeparation, Accumulator, Change);
                }
                Accumulator.Add(Change);
            } finally {
                //@ Unsupported property or method(C): 'Free'
                Change.Free;
            }
        }

    } // end TFormationBoid

    // TFormationBoid
    public class TFormationFlock: TFlock, 
    {
        // Published declarations
        public TFormationFlockCollection Boids
        {
          get {
            return GetBoids();
          }
          set {
            SetBoids(value);
          }
        }
        public TVector DesiredPosition
        {
          get {
            return FDesiredPosition;
          }
          set {
            SetDesiredPosition(value);
          }
        }
        public TVector DesiredVelocity
        {
          get {
            return FDesiredVelocity;
          }
          set {
            SetDesiredVelocity(value);
          }
        }
        public TFlockFormation Formation
        {
          get {
            return FFormation;
          }
          set {
            SetFormation(value);
          }
        }
        public TNotifyEvent OnDefineFormation
        {
          get {
            return FOnDefineFormation;
          }
          set {
            FOnDefineFormation = value;
          }
        }
        // Private declarations
        private TFlockFormation FFormation = TFlockFormation.ffLine;
        private TNotifyEvent FOnDefineFormation = null;
        private int FUpdateCount = 0;
        private TVector FDesiredPosition = null;
        private TVector FDesiredVelocity = null;
        // TFormationFlock
        public virtual void BeginFormationUpdate()
        {
            FUpdateCount ++;
        }

        //Constructor  Create( AOwner)
        public TFormationFlock(Component AOwner) : base(AOwner)
        {
            // turn down strength separation in relation to
            // cohesion
            this.StrengthAlignment = 1;
            this.StrengthAvoidance = 1;
            this.StrengthSeparation = 0.2;
            this.SeparationDistance = 1;
            FDesiredPosition = new TVector();
            FDesiredVelocity = new TVector();
            Formation = TFlockFormation.ffLine;
        }
        public override TFlockCollection CreateFlock()
        {
            TFlockCollection result;
            result = new TFormationFlockCollection(this, TFormationBoid);
            return result;
        }

        public virtual void DefineFormation()
        {
            //@ Unsupported property or method(C): 'Count'
            if ((Boids == null) || (Boids.Count == 0))
            {
                return;
            }
            switch(Formation)
            {
                case TFlockFormation.ffUserDefined:
                    // set boids to formation
                    // user defined formation... if event not assigned then just normal
                    // flocking behavior
                    if ((FOnDefineFormation != null))
                    {
                        OnDefineFormation[this];
                    }
                    break;
                case TFlockFormation.ffLine:
                    DefineLineFormation();
                    break;
                case TFlockFormation.ffSquare:
                    DefineSquareFormation();
                    break;
                case TFlockFormation.ffTwoSquares:
                    DefineTwoSquaresFormation();
                    break;
                case TFlockFormation.ffWedge:
                    DefineWedgeFormation();
                    break;
            }
        }

        public virtual void DefineLineFormation()
        {
            int i;
            float Sep;
            float s;
            TVector v;
            // Get the orthogonal vector to the group's orientation
            // Get the angle
            //@ Unsupported property or method(D): 'Z'
            //@ Unsupported property or method(D): 'X'
            s = (Math.Atan2(Convert.ToDouble(this.Velocity.Z), Convert.ToDouble(this.Velocity.X)) * ( 180  / Math.PI ));
            // add 90 to angle
            s = s + 90;
            v = new TVector(Math.Cos((s * ( Math.PI  / 180 ))), 0, Math.Sin((s * ( Math.PI  / 180 ))));
            try {
                //@ Unsupported property or method(C): 'Count'
                Sep = Units.FormationFlock.SeparationMagnitude * this.SeparationDistance * (Boids.Count / 2);
                // use separation distance
                //@ Unsupported property or method(C): 'Count'
                for (i = 0; i < Boids.Count; i ++ )
                {
                    //@ Unsupported property or method(C): 'Offset'
                    //@ Unsupported property or method(D): 'X'
                    Boids[i].Offset.X = v.X * Sep;
                    //@ Unsupported property or method(C): 'Offset'
                    //@ Unsupported property or method(D): 'Z'
                    Boids[i].Offset.Z = v.Z * Sep;
                    Sep = Sep - Units.FormationFlock.SeparationMagnitude * this.SeparationDistance;
                }
            } finally {
                //@ Unsupported property or method(C): 'Free'
                v.Free;
            }
        }

        protected void DefineSquare(int Index, int Count, TVector OffsetPos, TVector Vel, int Cols)
        {
            int Row;
            int Col;
            int i;
            float SepCol;
            float SepRow;
            TVector v;
            float s;
            // Get the orthogonal vector to the group's orientation
            // Get the angle
            s = (Math.Atan2(Vel.Z, Vel.X) * ( 180  / Math.PI ));
            // add 90 to angle
            s = s + 90;
            v = new TVector(Math.Cos((s * ( Math.PI  / 180 ))), 0, Math.Sin((s * ( Math.PI  / 180 ))));
            try {
                v.Normalize();
                SepCol = Units.FormationFlock.SeparationMagnitude * this.SeparationDistance * (Cols / 2);
                Row = 0;
                SepRow = Row * SepCol;
                Col = 0;
                // use separation distance
                for (i = Index; i < Count; i ++ )
                {
                    //@ Unsupported property or method(C): 'Offset'
                    //@ Unsupported property or method(D): 'X'
                    Boids[i].Offset.X = OffsetPos.X + v.X * SepCol - Vel.X * SepRow;
                    //@ Unsupported property or method(C): 'Offset'
                    //@ Unsupported property or method(D): 'Z'
                    Boids[i].Offset.Z = OffsetPos.Z + v.Z * SepCol - Vel.Z * SepRow;
                    Col ++;
                    if (Col > Cols)
                    {
                        Col = 0;
                        Row ++;
                        SepCol = Units.FormationFlock.SeparationMagnitude * this.SeparationDistance * (Cols / 2);
                        SepRow = Row * SepCol;
                    }
                    else
                    {
                        SepCol = SepCol - Units.FormationFlock.SeparationMagnitude * this.SeparationDistance;
                    }
                }
            } finally {
                //@ Unsupported property or method(C): 'Free'
                v.Free;
            }
        }

        public virtual void DefineSquareFormation()
        {
            int Cols;
            TVector OffsetPos;
            //@ Unsupported property or method(C): 'Count'
            Cols = Math.Round(Math.Sqrt(Convert.ToDouble(Boids.Count)));
            OffsetPos = new TVector();
            try {
                //@ Unsupported property or method(C): 'Count'
                DefineSquare(0, Boids.Count, OffsetPos, this.Velocity, Cols);
            } finally {
                //@ Unsupported property or method(C): 'Free'
                OffsetPos.Free;
            }
        }

        public virtual void DefineTwoSquaresFormation()
        {
            int Cols;
            TVector OffsetPos;
            TVector v;
            float s;
            // Get the orthogonal vector to the group's orientation
            // Get the angle
            //@ Unsupported property or method(D): 'Z'
            //@ Unsupported property or method(D): 'X'
            s = (Math.Atan2(Convert.ToDouble(this.Velocity.Z), Convert.ToDouble(this.Velocity.X)) * ( 180  / Math.PI ));
            // add 90 to angle
            s = s + 90;
            v = new TVector(Math.Cos((s * ( Math.PI  / 180 ))), 0, Math.Sin((s * ( Math.PI  / 180 ))));
            //@ Unsupported property or method(C): 'Count'
            Cols = Math.Round(Math.Sqrt(Convert.ToDouble(Boids.Count / 2)));
            OffsetPos = new TVector();
            try {
                v.SetMagnitudeOfVector(1);
                OffsetPos.SetXYZ(v.X, 0, v.Z);
                OffsetPos.SetMagnitudeOfVector(10);
                //@ Unsupported property or method(C): 'Count'
                DefineSquare(0, Boids.Count / 2, OffsetPos, this.Velocity, Cols);
                OffsetPos.SetXYZ(v.X, 0, v.Z);
                OffsetPos.SetMagnitudeOfVector( -10);
                //@ Unsupported property or method(C): 'Count'
                //@ Unsupported property or method(C): 'Count'
                DefineSquare((Boids.Count / 2), Boids.Count, OffsetPos, this.Velocity, Cols);
            } finally {
                //@ Unsupported property or method(C): 'Free'
                OffsetPos.Free;
                //@ Unsupported property or method(C): 'Free'
                v.Free;
            }
        }

        public virtual void DefineWedgeFormation()
        {
            int i;
            float Sep;
            float s1;
            float s2;
            TVector v1;
            TVector v2;
            // Get the 120 Degrees vector to the group's orientation
            // Get the angle
            //@ Unsupported property or method(D): 'Z'
            //@ Unsupported property or method(D): 'X'
            s1 = (Math.Atan2(Convert.ToDouble(this.Velocity.Z), Convert.ToDouble(this.Velocity.X)) * ( 180  / Math.PI ));
            // add 120 to angle
            s2 = s1 - 120;
            s1 = s1 + 120;
            v1 = new TVector(Math.Cos((s1 * ( Math.PI  / 180 ))), 0, Math.Sin((s1 * ( Math.PI  / 180 ))));
            v2 = new TVector(Math.Cos((s2 * ( Math.PI  / 180 ))), 0, Math.Sin((s2 * ( Math.PI  / 180 ))));
            try {
                //@ Unsupported property or method(C): 'Count'
                Sep = Units.FormationFlock.SeparationMagnitude * this.SeparationDistance * (Boids.Count / 2);
                // use separation distance
                //@ Unsupported property or method(C): 'Count'
                for (i = 0; i <= (Boids.Count / 2); i ++ )
                {
                    //@ Unsupported property or method(C): 'Offset'
                    //@ Unsupported property or method(D): 'X'
                    Boids[i].Offset.X = v1.X * Sep;
                    //@ Unsupported property or method(C): 'Offset'
                    //@ Unsupported property or method(D): 'Z'
                    Boids[i].Offset.Z = v1.Z * Sep;
                    //@ Unsupported property or method(C): 'Count'
                    //@ Unsupported property or method(C): 'OffSet'
                    //@ Unsupported property or method(D): 'X'
                    Boids[Boids.Count - 1 - i].OffSet.X = v2.X * Sep;
                    //@ Unsupported property or method(C): 'Count'
                    //@ Unsupported property or method(C): 'OffSet'
                    //@ Unsupported property or method(D): 'Z'
                    Boids[Boids.Count - 1 - i].OffSet.Z = v2.Z * Sep;
                    Sep = Sep - Units.FormationFlock.SeparationMagnitude * this.SeparationDistance;
                }
                //@ Unsupported property or method(C): 'Count'
                //@ Unsupported property or method(C): 'Count'
                if ((Boids.Count > 1) && ((Boids.Count % 2) != 0))
                {
                    //@ Unsupported property or method(C): 'Count'
                    //@ Unsupported property or method(C): 'Offset'
                    //@ Unsupported property or method(D): 'X'
                    Boids[Boids.Count / 2 + 1].Offset.X = 0;
                    //@ Unsupported property or method(C): 'Count'
                    //@ Unsupported property or method(C): 'Offset'
                    //@ Unsupported property or method(D): 'Z'
                    Boids[Boids.Count / 2 + 1].Offset.Z = 0;
                }
            } finally {
                //@ Unsupported property or method(C): 'Free'
                v1.Free;
                //@ Unsupported property or method(C): 'Free'
                v2.Free;
            }
        }

        public virtual void EndFormationUpdate()
        {
            FUpdateCount -= 1;
            FormationChange();
        }

        protected void FormationChange()
        {
            if (FUpdateCount == 0)
            {
                DefineFormation();
            }
        }

        private TFormationFlockCollection GetBoids()
        {
            TFormationFlockCollection result;
            //@ Unsupported property or method(D): 'Boids'
            result = ((base.Boids) as TFormationFlockCollection);
            return result;
        }

        private void SetBoids(TFormationFlockCollection Value)
        {
            //@ Unsupported property or method(D): 'Boids'
            base.Boids = Value;
        }

        private void SetFormation(TFlockFormation Value)
        {
            if (Formation != Value)
            {
                FFormation = Value;
                DefineFormation();
            }
        }

        public override void Execute()
        {
            int i;
        }

        public void Assign(Object Source)
        {
        }

        //@ Destructor  Destroy()
        ~TFormationFlock()
        {
        }
        public void SetDesiredPosition(TVector Value)
        {
        }

        public void SetDesiredVelocity(TVector Value)
        {
        }

        public void VectorChange(TVector Sender)
        {
        }

        public void SetSeparationDistance(float Value)
        {
        }

    } // end TFormationFlock

    public class TFormationFlockCollection: TFlockCollection
    {
        // Protected declarations
        protected TFormationFlock Flock
        {
          get {
            return GetFlock();
          }
        }
        // Public declarations
        public TFormationBoid this[int Index]
        {
          get {
            return GetItem(Index);
          }
          set {
            SetItem(Index, value);
          }
        }
        public TFlockCollection(TCustomFlock Flock) : base(Flock)
        {
        }
        public TFlockCollection(TCustomFlock Flock, TBoidClass BoidClass) : base(Flock, BoidClass)
        {
        }
        // TFormationFlockCollection
        private TFormationFlock GetFlock()
        {
            TFormationFlock result;
            result = ((base.Flock) as TFormationFlock);
            return result;
        }

        // Private declarations
        private TFormationBoid GetItem(int Index)
        {
            TFormationBoid result;
            result = ((base[Index]) as TFormationBoid);
            return result;
        }

        private void SetItem(int Index, TFormationBoid Value)
        {
            base[Index] = Value;
        }

    } // end TFormationFlockCollection

    // TFormationFlockCollection
    public enum TFlockFormation
    {
        ffUserDefined,
        ffLine,
        ffSquare,
        ffTwoSquares,
        ffWedge
    } // end TFlockFormation

}

namespace FormationFlock.Units
{
    public class FormationFlock
    {
        public const int SeparationMagnitude = 3;
        // TFormationFlock
        public static void Register()
        {
            //@ Unsupported function or procedure: 'RegisterComponents'
            RegisterComponents("AI", new TFormationFlock[] {TFormationFlock});
        }

    } // end FormationFlock

}

